/*
 * Copyright (c) 2007, Solido Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of Solido Systems nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 :* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package tuplesoup.core;

import TransactionalIO.core.TransactionalFile;
import TransactionalIO.interfaces.IOOperations;
import java.io.*;
import java.util.*;
import java.nio.channels.*;
import tuplesoup.filter.*;
import dstm2.Defaults;
import dstm2.SpecialTransactionalFile;
import dstm2.atomic;
import dstm2.Thread;
import dstm2.util.HashMap2;
import dstm2.factory.Factory;
import java.util.concurrent.Callable;

/**
 * The table stores a group of rows.
 * Every row must have a unique id within a table.
 */
public class DualFileTable implements Table {

    private int INDEXCACHESIZE = 8192;
    private IOOperations[] filestream = new IOOperations[8];
    private IOOperations[] filerandom = new IOOperations[8];

    private TableIndex index = null;
    private boolean rowswitch = true;
    private String title;
    private String location;
    private HashMap2<TableIndexNode> indexcache;
    static Factory<TableIndexNode> tableindexnodefactory = Thread.makeFactory(TableIndexNode.class);

    /**
     * Create a new table object with a specific index model
     */
    public DualFileTable(String title, String location, int indextype) throws IOException {

        this.title = title;
        this.location = location;
        if (!this.location.endsWith(File.separator)) {
            this.location += File.separator;
        }
        switch (indextype) {
            case PAGED:
                index = new PagedIndex(getFileName(INDEX));
                break;

        }
        indexcache = new HashMap2<TableIndexNode>();
        openFiles();
	openUpdateFiles();
    }

    /**
     * Close all open file streams
     */
    private IOOperations getFile(boolean in_file_update, int type){
	 if (in_file_update)
		return filerandom[type];
	else 
		return filestream[type];
    }
    
    
    public void close() {
        try {
	    for (int i=0; i<filerandom.length; i++){
		if (filerandom[i] != null)
			filerandom[i].close();
		if (filestream[i] != null)
			filestream[i].close();
	    }
            index.close();
        } catch (Exception e) {
        }
    }

    /** 
     * Returns the name of this table
     */
    public String getTitle() {
        return title;
    }

    /**
     * Returns the location of this tables datafiles
     */
    public String getLocation() {
        return location;
    }

    protected String getFileName(int type) {
        switch (type) {
            case INDEX:
                return location + title + ".index";
            case FILEB:
                return location + title + ".b";
            case FILEA:
                return location + title + ".a";
            case FILEC:
                return location + title + ".c";
            case FILED:
                return location + title + ".d";
            case FILEE:
                return location + title + ".e";
            case FILEF:
                return location + title + ".f";
            case FILEG:
                return location + title + ".g";
            case FILEH:
                return location + title + ".h";


        }
        return null;
    }

    /**
     * Delete the files created by this table object.
     * Be aware that this will delete any data stored in this table!
     */
    public void deleteFiles() {
        try {
            File ftest = new File(getFileName(FILEA));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEB));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEC));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILED));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEE));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEF));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEG));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(FILEH));
            ftest.delete();
        } catch (Exception e) {
        }
        try {
            File ftest = new File(getFileName(INDEX));
            ftest.delete();
        } catch (Exception e) {
        }
    }

    private void openFiles() throws IOException {
        try{
        for (int i=0; i<filestream.length;i++){
	        if (filestream[i] == null) {
                    if (Defaults.INEVITABLE) {
                        filestream[i] = new SpecialTransactionalFile(getFileName(i), "rw");
                    } else {
                        filestream[i] = new TransactionalFile(getFileName(i), "rw");
                    }
                    File ftest = new File(getFileName(i));
                    filestream[i].seek(ftest.length());
       		}
        }
	}
	catch(IOException ex){
		ex.printStackTrace();
	}
   }
   

    public void openUpdateFiles(){
	try{
        for (int i=0; i<filerandom.length;i++){
	        if (filerandom[i] == null) 
                    if (Defaults.INEVITABLE) {
                        filerandom[i] = new SpecialTransactionalFile(getFileName(i), "rw");
                    } else {
                        filerandom[i] = new TransactionalFile(getFileName(i), "rw");
                    }
        }
	}
	catch(IOException ex){
		ex.printStackTrace();
	}
    }

   public void setIndexCacheSize(int size){
	INDEXCACHESIZE = size;
   } 

    private void addCacheEntry(TableIndexEntry entry) {
        TableIndexNode node = tableindexnodefactory.create();
        node.setData(entry);
        indexcache.put(entry.getId().hashCode(), node);
    }

    private void updateCacheEntry(TableIndexEntry entry){
	if (indexcache.containsKey(entry.getId().hashCode())){
		TableIndexNode node = indexcache.get(entry.getId().hashCode());
		node.setData(entry);
	//		System.out.println(Thread.currentThread() + " update id " + entry.getId());
	}
	else
		addCacheEntry(entry);	
    }

    private void removeCacheEntry(String id) {
	if (indexcache.containsKey(id.hashCode())){
		TableIndexNode node = indexcache.get(id.hashCode());
		indexcache.remove(id.hashCode());
	}
    }

    private TableIndexEntry getCacheEntry(final String id) {
        if (indexcache.containsKey(id.hashCode())) {
            TableIndexNode node = indexcache.get(id.hashCode());
            return node.getData();
        } else {
            return null;
        }
    }


   public void addOrUpdateRow(Row row) throws IOException{
	Row tmprow = getRow(row.getId());
	if (tmprow == null){
		addRow(row);
	}
	else{
		updateRow(row);
	}

   }	

        // Distribute new rows between the two datafiles by using the rowswitch, but don't spend time synchronizing... this does not need to be acurate!
    public void addRow(Row row) throws IOException {
//        int type = (int)(Math.random()*8);
        addRowtoFiles(row, Integer.valueOf(row.getId())%8/*Integer.valueOf(Thread.currentThread().getName().substring(7))*/);
    }

    private void addRowtoFiles(final Row row, final int file){
        Thread.doIt(new Callable<Boolean>() {
            public Boolean call() throws IOException {
                int pre = (int) getFile(false, file).getFilePointer();
                row.writeToStream(getFile(false, file));
                index.addEntry(row.getId(), row.getSize(), file, pre);
                if (INDEXCACHESIZE > 0) {
                    TableIndexEntryImplementation entry = new TableIndexEntryImplementation(row.getId(), row.getSize(), file, pre);
                    addCacheEntry(entry.entry);
                }
                return true;
            }
        });
    }

    public void updateRow(final Row row) throws IOException{
         TableIndexEntry entry=null;
         // Handle index entry caching
         if(INDEXCACHESIZE>0){
            entry = Thread.doIt(new Callable<TableIndexEntry>() {
                public TableIndexEntry call() throws IOException {
	                 TableIndexEntry entry=getCacheEntry(row.getId());
        	         if(entry==null){
                	     entry=index.scanIndex(row.getId());
	                     addCacheEntry(entry);
        	         }
			 return entry;
             }
	   });
	 }
         else{
            entry = Thread.doIt(new Callable<TableIndexEntry>() {
                public TableIndexEntry call() throws IOException {
            		 TableIndexEntry entry=index.scanIndex(row.getId());
			 return entry;
		}
	    });
         }
	 final int location = entry.getLocation();
	 final long position = entry.getPosition();
	 if (location != DELETE){	
	         if(entry.getRowSize()>=row.getSize()){
        	    // Add to the existing location
                	Thread.doIt(new Callable<Boolean>() {
				  public Boolean call() throws IOException {
	                        	        filerandom[location].seek(position);
        	                        	row.writeToStream(filerandom[location]);
						return true;
        	        	  }
	        	});
            
		 }
		else {	
			updateRowinFiles(row, location);
		}
	}


    }

    public void updateRowinFiles(final Row row, final int location/*, final long position*/){
        Thread.doIt(new Callable<Boolean>() {
	      	public Boolean call() throws IOException {
	              int pre=(int)filestream[location].getFilePointer();
        	      row.writeToStream(filestream[location]);
	              index.updateEntry(row.getId(),row.getSize(),location,pre);

              // Handle index entry caching
        	      if(INDEXCACHESIZE>0){
                	  TableIndexEntryImplementation entry = new TableIndexEntryImplementation(row.getId(),row.getSize(), location, pre);
                	  updateCacheEntry(entry.entry);
	              }
		      return true;
    		}	
	});
    }

    public void deleteRow(final Row row){
	if (INDEXCACHESIZE > 0)
		Thread.doIt(new Callable<Boolean>(){
			public Boolean call(){
				removeCacheEntry(row.getId());
				return true;
			}
		});
	Thread.doIt(new Callable<Boolean>(){
		public Boolean call(){
			try{
			        index.updateEntry(row.getId(), row.getSize(), DELETE, 0);
			} catch (IOException ex){
				ex.printStackTrace();
			}
			return true;
		}
	});
    } 
    /**
     * Returns a single row stored in this table.
     * If the row does not exist in the table, null will be returned.
     */
    public Row getRow(final String id) throws IOException {
        TableIndexEntry entry = null;
        // Handle index entry caching
        if (INDEXCACHESIZE > 0) {
            entry = Thread.doIt(new Callable<TableIndexEntry>() {
                public TableIndexEntry call() throws IOException {
                    TableIndexEntry entry = getCacheEntry(id);
                    if (entry == null) {
                        entry = index.scanIndex(id);
                        if (entry != null) {
                            addCacheEntry(entry);
                        }
                    }
		    return entry;	
                }
            });
        } else {
            entry = Thread.doIt(new Callable<TableIndexEntry>() {
                public TableIndexEntry call() throws IOException {
            		TableIndexEntry entry = index.scanIndex(id);
		        return entry;
		}
            });
        }
        if (entry != null) {
            long pos = entry.getPosition();
	    int loc = entry.getLocation();
            long dataoffset = 0;
            DataInputStream data = null;
            if (loc != DELETE)
	            data = new DataInputStream(new BufferedInputStream(new FileInputStream(getFileName(loc))));
            
            if (data != null) {
                while (dataoffset != pos) {
                    dataoffset += data.skipBytes((int) (pos - dataoffset));
                }
		Row row = null;
		try{
                row = Row.readFromStream(data);	
                data.close();
                return row;
		}catch(EOFException e){
                	data.close();
			return null;
		}
            }

        }
        return null;
        //return false;
    }

    public TupleStream getRows(List<String> rows) throws IOException {
	return null;
    }

    public TupleStream getRows(RowMatcher matcher) throws IOException {
	return null;
    }

    public TupleStream getRows(List<String> rows, RowMatcher matcher) throws IOException {
	return null;
    }

    public TupleStream getRows() throws IOException {
	return null;
    }
}
